from pyramid.events import NewRequest
from pyramid.httpexceptions import HTTPBadRequest

from pyramid_restler.view import RESTfulView


def add_restful_routes(self, name, factory, view=RESTfulView,
                       route_kw=None, view_kw=None):
    """Add a set of RESTful routes for an entity.

    URL patterns for an entity are mapped to a set of views encapsulated in
    a view class. The view class interacts with the model through a context
    adapter that knows the particulars of that model.

    To use this directive in your application, first call
    `config.include('pyramid_restler')` somewhere in your application's
    `main` function, then call `config.add_restful_routes(...)`.

    ``name`` is used as the base name for all route names and patterns. In
    route names, it will be used as-is. In route patterns, underscores will
    be converted to dashes.

    ``factory`` is the model adapter that the view interacts with. It can be
    any class that implements the :class:`pyramid_restler.interfaces.IContext`
    interface.

    ``view`` must be a view class that implements the
    :class:`pyramid_restler.interfaces.IView` interface.

    Additional route and view keyword args can be passed directly through to
    all `add_route` and `add_view` calls. Pass ``route_kw`` and/or ``view_kw``
    as dictionaries to do so.

    """
    route_kw = {} if route_kw is None else route_kw
    view_kw = {} if view_kw is None else view_kw

    subs = dict(
        name=name,
        slug=name.replace('_', '-'),
        id='{id}',
        renderer='{renderer}')

    def add_route(name, pattern, attr, method):
        name = name.format(**subs)
        pattern = pattern.format(**subs)
        self.add_route(
            name, pattern, factory=factory,
            request_method=method, **route_kw)
        self.add_view(
            view=view, attr=attr, route_name=name,
            request_method=method, **view_kw)

    # Get collection
    add_route(
        'get_{name}_collection_rendered', '/{slug}.{renderer}',
        'get_collection', 'GET')
    add_route(
        'get_{name}_collection', '/{slug}', 'get_collection', 'GET')

    # Get member
    add_route(
        'get_{name}_rendered', '/{slug}/{id}.{renderer}', 'get_member', 'GET')
    add_route('get_{name}', '/{slug}/{id}', 'get_member', 'GET')

    # Create member
    add_route('create_{name}', '/{slug}', 'create_member', 'POST')

    # Update member
    add_route('update_{name}', '/{slug}/{id}', 'update_member', 'PUT')

    # Delete member
    add_route('delete_{name}', '/{slug}/{id}', 'delete_member', 'DELETE')


def enable_POST_tunneling(self, allowed_methods=('PUT', 'DELETE')):
    """Allow other request methods to be tunneled via POST.

    This allows PUT and DELETE requests to be tunneled via POST requests.
    The method can be specified using a parameter or a header...

    The name of the parameter is '$method'; it can be a query or POST
    parameter. The query parameter will be preferred if both the query and
    POST parameters are present in the request.

    The name of the header is 'X-HTTP-Method-Override'. If the parameter
    described above is passed, this will be ignored.

    The request method will be overwritten before it reaches application
    code, such that the application will never be aware of the original
    request method. Likewise, the parameter and header will be removed from
    the request, and the application will never see them.

    """
    param_name = '$method'
    header_name = 'X-HTTP-Method-Override'
    allowed_methods = set(allowed_methods)
    disallowed_message = (
        'Only these methods may be tunneled over POST: {0}.'
        .format(sorted(list(allowed_methods))))
    def new_request_subscriber(event):
        request = event.request
        if request.method == 'POST':
            if param_name in request.GET:
                method = request.GET[param_name]
            elif param_name in request.POST:
                method = request.POST[param_name]
            elif header_name in request.headers:
                method = request.headers[header_name]
            else:
                return  # Not a tunneled request
            if method in allowed_methods:
                request.GET.pop(param_name, None)
                request.POST.pop(param_name, None)
                request.headers.pop(header_name, None)
                request.method = method
            else:
                raise HTTPBadRequest(disallowed_message)
    self.add_subscriber(new_request_subscriber, NewRequest)



import json
from unittest import TestCase

from pyramid.config import Configurator
from pyramid.events import NewRequest
from pyramid.httpexceptions import HTTPBadRequest, HTTPNotFound
from pyramid.response import Response
from pyramid.testing import DummyRequest

from webob.request import MIMEAccept

try:
    import sqlalchemy
except ImportError:
    pass
else:
    from sqlalchemy.engine import create_engine
    from sqlalchemy.ext.declarative import declarative_base
    from sqlalchemy.orm import Session
    from sqlalchemy.schema import Column
    from sqlalchemy.types import Integer, String

from zope.interface import implements

from pyramid_restler.interfaces import IContext
from pyramid_restler.model import SQLAlchemyORMContext
from pyramid_restler.view import RESTfulView


class Test_SQLAlchemyORMContext(TestCase):

    def setUp(self):
        # For each test, create a new database and populate it with a few
        # things. Then create a context instance to be used by the test.
        engine = create_engine('sqlite://')
        session = Session(bind=engine)
        Base = declarative_base()
        class Entity(Base):
            __tablename__ = 'entity'
            id = Column(Integer, primary_key=True)
            value = Column(String)
        Base.metadata.create_all(bind=engine)
        session.add_all([
            Entity(id=1, value='one'),
            Entity(id=2, value='two'),
            Entity(id=3, value='three'),
        ])
        session.commit()
        class ContextFactory(SQLAlchemyORMContext):
            entity = Entity
        request = DummyRequest()
        request.db_session = session
        self.context = ContextFactory(request)

    def test_interface(self):
        self.assertTrue(IContext.implementedBy(SQLAlchemyORMContext))
        self.assertTrue(IContext.providedBy(self.context))

    def test_get_collection(self):
        collection = self.context.get_collection()
        self.assertEqual(3, len(collection))

    def test_get_member(self):
        member = self.context.get_member(1)
        self.assertEqual(member.id, 1)
        self.assertEqual(member.value, 'one')

    def test_get_nonexistent_member(self):
        member = self.context.get_member(42)
        self.assert_(member is None)

    def test_create_member(self):
        member = self.context.create_member(dict(value='four'))
        member = self.context.session.query(self.context.entity).get(member.id)
        self.assertEqual(member, member)

    def test_update_member(self):
        self.context.update_member(1, dict(value='ONE'))
        member = self.context.get_member(1)
        self.assertEqual('ONE', member.value)

    def test_delete_member(self):
        member = self.context.get_member(1)
        self.assert_(member is not None)
        self.context.delete_member(1)
        member = self.context.get_member(1)
        self.assert_(member is None)

    def test_collection_to_json(self):
        collection = self.context.get_collection()
        json_collection = self.context.to_json(collection)
        self.assertTrue(isinstance(json_collection, basestring))
        should_equal = [
            {'id': 1, 'value': 'one'},
            {'id': 2, 'value': 'two'},
            {'id': 3, 'value': 'three'},
        ]
        self.assertEqual(json.loads(json_collection)['results'], should_equal)

    def test_member_to_json(self):
        member = self.context.get_member(1)
        json_member = self.context.to_json(member)
        self.assertTrue(isinstance(json_member, basestring))
        should_equal = {'id': 1, 'value': 'one'}
        self.assertEqual(json.loads(json_member)['results'], should_equal)

    def test_get_member_id_as_string(self):
        member = self.context.get_member(1)
        id = self.context.get_member_id_as_string(member)
        self.assertEqual(id, '1')


class Test_RESTfulView(TestCase):

    def test_get_collection(self):
        request = DummyRequest(path='/thing.json')
        request.matchdict = {'renderer': 'json'}
        view = RESTfulView(_dummy_context_factory(), request)
        response = view.get_collection()
        self.assertTrue(isinstance(response, Response))

    def test_get_member(self):
        request = DummyRequest(path='/thing/1.json')
        request.matchdict = {'id': 1, 'renderer': 'json'}
        view = RESTfulView(_dummy_context_factory(), request)
        response = view.get_member()
        self.assertTrue(isinstance(response, Response))
        self.assertEqual(response.status_int, 200)

    def test_create_member(self):
        request = DummyRequest(
            path='/thing', post={'val': 'four'},
            content_type='application/x-www-form-urlencoded')
        view = RESTfulView(_dummy_context_factory(), request)
        response = view.create_member()
        self.assertTrue(isinstance(response, Response))
        self.assertEqual(response.status_int, 201)

    def test_update_member(self):
        request = DummyRequest(
            method='PUT', path='/thing/1', post={'val': 'ONE'},
            content_type='application/x-www-form-urlencoded')
        request.matchdict = {'id': 1}
        view = RESTfulView(_dummy_context_factory(), request)
        response = view.update_member()
        self.assertTrue(isinstance(response, Response))
        self.assertEqual(response.status_int, 204)

    def test_delete_member(self):
        request = DummyRequest(method='DELETE', path='/thing/1',)
        request.matchdict = {'id': 1, 'renderer': 'json'}
        view = RESTfulView(_dummy_context_factory(), request)
        response = view.delete_member()
        self.assertTrue(isinstance(response, Response))
        self.assertEqual(response.status_int, 204)
        self.assertRaises(HTTPNotFound, view.get_member)

    def test_get_nonexistent_member_should_raise_404(self):
        request = DummyRequest(path='/thing/42.json')
        request.matchdict = {'id': 42, 'renderer': 'json'}
        view = RESTfulView(_dummy_context_factory(), request)
        self.assertRaises(HTTPNotFound, view.get_member)

    def test_get_member_specific_fields(self):
        request = DummyRequest(path='/thing/1.json', params={'$fields': '["id"]'})
        request.matchdict = {'id': 1, 'renderer': 'json'}
        view = RESTfulView(_dummy_context_factory(), request)
        response = view.get_member()
        member = json.loads(response.body)['results'][0]
        self.assert_(member.keys() == ['id'])

    def test_wrapped_response(self):
        request = DummyRequest(path='/thing/1.json', params={'$wrap': 'true'})
        request.matchdict = {'id': 1, 'renderer': 'json'}
        view = RESTfulView(_dummy_context_factory(), request)
        self.assertTrue(view.wrap)
        response = view.get_member()
        content = json.loads(response.body)
        self.assert_('results' in content)
        member = content['results'][0]
        self.assert_('id' in member and 'val' in member)

    def test_unwrapped_response(self):
        request = DummyRequest(path='/thing/1.json', params={'$wrap': 'false'})
        request.matchdict = {'id': 1, 'renderer': 'json'}
        view = RESTfulView(_dummy_context_factory(), request)
        self.assertFalse(view.wrap)
        response = view.get_member()
        member = json.loads(response.body)[0]
        self.assert_('id' in member and 'val' in member)

    def test_xml_renderer(self):
        request = DummyRequest(path='/thing/1')
        request.accept = MIMEAccept('Accept', 'application/xml')
        request.matchdict = {'id': 1}
        view = RESTfulView(_dummy_context_factory(), request)
        self.assertRaises(HTTPBadRequest, view.get_member)

    def test_unknown_renderer_should_raise_400(self):
        request = DummyRequest(path='/thing/1.xyz')
        request.matchdict = {'id': 1, 'renderer': 'xyz'}
        view = RESTfulView(_dummy_context_factory(), request)
        self.assertRaises(HTTPBadRequest, view.get_member)


class Test_add_restful_routes(TestCase):

    def _make_config(self, autocommit=True, add_view=None):
        config = Configurator(autocommit=autocommit)
        if add_view is not None:
            config.add_view = add_view
        config.include('pyramid_restler')
        return config

    def _make_add_view(self):
        class AddView(object):
            def __init__(self):
                self.views = []
            def __call__(self, **kw):
                self.views.append(kw)
            def count(self):
                return len(self.views)
        return AddView()

    def test_directive_registration(self):
        config = self._make_config()
        self.assertTrue(hasattr(config, 'add_restful_routes'))

    def test_add_restful_routes(self):
        config = self._make_config(add_view=self._make_add_view())
        config.add_restful_routes('thing', _dummy_context_factory())
        self.assertEqual(7, config.add_view.count())


class Test_POST_tunneling(TestCase):

    def _make_app(self):
        config = Configurator()
        config.include('pyramid_restler')
        config.enable_POST_tunneling()
        return config.make_wsgi_app()

    def test_POST_without_tunnel(self):
        app = self._make_app()
        request = DummyRequest(method='POST')
        self.assertEqual('POST', request.method)
        self.assert_('$method' not in request.params)
        app.registry.notify(NewRequest(request))
        self.assertEqual('POST', request.method)
        self.assert_('$method' not in request.params)

    def _assert_before(self, request):
        self.assertEqual('POST', request.method)

    def _assert_after(self, request, method):
        self.assertEqual(request.method, method)
        self.assert_('$method' not in request.GET)
        self.assert_('$method' not in request.POST)
        self.assert_('X-HTTP-Method-Override' not in request.headers)

    def test_PUT_using_GET_param(self):
        app = self._make_app()
        request = DummyRequest(method='POST')
        request.GET = {'$method': 'PUT'}
        request.POST = {'$method': 'DUMMY'}
        request.headers['X-HTTP-Method-Override'] = 'DUMMY'
        self._assert_before(request)
        app.registry.notify(NewRequest(request))
        self._assert_after(request, 'PUT')

    def test_PUT_using_POST_param(self):
        app = self._make_app()
        request = DummyRequest(method='POST')
        request.POST = {'$method': 'PUT'}
        request.headers['X-HTTP-Method-Override'] = 'DUMMY'
        self._assert_before(request)
        app.registry.notify(NewRequest(request))
        self._assert_after(request, 'PUT')

    def test_PUT_using_header(self):
        app = self._make_app()
        request = DummyRequest(method='POST')
        request.headers['X-HTTP-Method-Override'] = 'PUT'
        self._assert_before(request)
        app.registry.notify(NewRequest(request))
        self._assert_after(request, 'PUT')

    def test_DELETE_using_POST_param(self):
        app = self._make_app()
        request = DummyRequest(method='POST')
        request.POST = {'$method': 'DELETE'}
        request.headers['X-HTTP-Method-Override'] = 'DUMMY'
        self._assert_before(request)
        app.registry.notify(NewRequest(request))
        self._assert_after(request, 'DELETE')

    def test_unknown_method_using_param(self):
        app = self._make_app()
        request = DummyRequest(params={'$method': 'PANTS'}, method='POST')
        self._assert_before(request)
        self.assertRaises(HTTPBadRequest, app.registry.notify, NewRequest(request))


def _dummy_context_factory():

    class Context(object):

        implements(IContext)

        _collection = [
            {'id': 1, 'val': 'one'},
            {'id': 2, 'val': 'two'},
            {'id': 3, 'val': 'three'},
        ]

        def _get_member_by_id(self, id):
            for m in self._collection:
                if m['id'] == id:
                    return m
            else:
                return None

        def get_collection(self):
            return self._collection

        def get_member(self, id):
            return self._get_member_by_id(id)

        def create_member(self, data):
            next_id = max(m['id'] for m in self._collection) + 1
            member = {'id': next_id, 'val': data['val']}
            return member

        def update_member(self, id, data):
            member = self._get_member_by_id(id)
            if member is None:
                return None
            for name in data:
                member[name] = data[name]
            return member

        def delete_member(self, id):
            for i, m in enumerate(self._collection):
                if m['id'] == id:
                    self._collection.pop(i)
            return None

        def get_member_id_as_string(self, member):
            return str(member['id'])

        def to_json(self, value, fields=None, wrap=True):
            if isinstance(value, dict):
                value = [value]
            if fields:
                for i, r in enumerate(value):
                    value[i] = dict((k, r[k]) for k in fields)
            if wrap:
                response = dict(results=value)
            else:
                response = value
            return json.dumps(response)

    return Context()



"""
SQLAlchemy ORM Example
======================

To run this example, first install pyramid_restler and SQLAlchemy (perhaps
in a virtualenv). Once that's done, cd into the examples/ directory, run
`python sqlalchemy_orm_example.py`, and open http://localhost:5000/ in
your browser. From there, you can play around with CRUD from a very simple
UI.

Note: a temporary SQLite database is created in the current working
directory every time this module is run; it is removed automatically when
the server is shut down normally (e.g., via Ctrl-C). The `DB_NAME` global
specifies the name of this database.

"""
import os

from paste.httpserver import serve

from pyramid.config import Configurator

from pyramid_restler.model import SQLAlchemyORMContext

from sqlalchemy.engine import create_engine
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker
from sqlalchemy.schema import Column
from sqlalchemy.types import Integer, String


DB_NAME = 'pyramid_restler_example.db'

engine = create_engine('sqlite:///{0}'.format(DB_NAME))
Session = sessionmaker(bind=engine)
Base = declarative_base(bind=engine)


class MyThing(Base):

    __tablename__ = 'my_thing'

    id = Column(Integer, primary_key=True)
    title = Column(String(100), nullable=False)
    description = Column(String)


class MyThingContextFactory(SQLAlchemyORMContext):

    entity = MyThing

    def session_factory(self):
        return Session()


def root_view(context, request):
    things = Session().query(MyThing).all()
    return dict(things=things, Thing=MyThing)


def main(**settings):
    create_and_populate_database()
    config = Configurator(settings=settings)
    config.add_route('root', '/')
    config.add_view(route_name='root', view=root_view, renderer='example.mako')
    config.include('pyramid_restler')
    config.add_restful_routes('thing', MyThingContextFactory)
    config.enable_POST_tunneling()
    return config.make_wsgi_app()


def create_and_populate_database():
    Base.metadata.create_all()
    session = Session()
    session.add_all([
        MyThing(title='One'),
        MyThing(title='Two'),
        MyThing(title='Three'),
    ])
    session.commit()


if __name__ == '__main__':
    settings = {'mako.directories': '.'}
    app = main(**settings)
    serve(app, host='0.0.0.0', port=5000)
    os.remove(DB_NAME)



from pyramid_restler.config import add_restful_routes, enable_POST_tunneling


def includeme(config):
    config.add_directive('add_restful_routes', add_restful_routes)
    config.add_directive('enable_POST_tunneling', enable_POST_tunneling)



from zope.interface import Attribute, Interface


class IView(Interface):

    def __init__(context, request):
        """Initialize view class."""

    def get_collection():
        """Get the entire collection.

        GET /entity -> 200 OK, list of members

        """

    def get_member():
        """Get a specific member by ID.

        GET /entity/id -> 200 OK, member

        """

    def create_member():
        """Create a new member.

        POST /entity?POST_data -> 201 Created, location of new member

        """

    def update_member(id):
        """Update an existing member.

        PUT /entity/id?POST_data -> 204 No Content

        """

    def delete_member(id):
        """Delete an existing member.

        DELETE /entity/id -> 204 No Content

        """

    def render_to_response(value, fields=None):
        """Render a member or list of members to an appropriate response.

        The request should include an indication of the client's preferred
        representation. This can be specified via the Accept header or by
        using the {.renderer} notation at the end of the URL path.

        """


class IContext(Interface):
    """Interface for adapting a model entity to a view context."""

    entity = Attribute('The entity to operate on.')

    default_fields = Attribute('A list of fields to include in the result.')

    def __init__(request):
        """Initialize context."""

    def get_collection():
        """Return the entire collection."""

    def get_member(id):
        """Return the member identified by ``id``."""

    def create_member(**data):
        """Create a new member."""

    def update_member(id, **data):
        """Update an existing member."""

    def delete_member(id):
        """Delete an existing member."""

    def get_member_id_as_string(member):
        """Get string representation of ``member`` ID."""



import json

from pyramid.decorator import reify
from pyramid.httpexceptions import HTTPBadRequest, HTTPNotFound
from pyramid.response import Response

from zope.interface import implements

from pyramid_restler.interfaces import IView


class RESTfulView(object):

    implements(IView)

    def __init__(self, context, request):
        self.context = context
        self.request = request

    def get_collection(self):
        collection = self.context.get_collection()
        return self.render_to_response(collection)

    def get_member(self):
        id = self.request.matchdict['id']
        member = self.context.get_member(id)
        return self.render_to_response(member)

    def _get_data(self):
        content_type = self.request.content_type
        if content_type == 'application/json':
            data = json.loads(self.request.body)
        elif content_type == 'application/x-www-form-urlencoded':
            data = dict(self.request.POST)
        return data

    def create_member(self):
        member = self.context.create_member(self._get_data())
        id = self.context.get_member_id_as_string(member)
        headers = {'Location': '/'.join((self.request.path, id))}
        return Response(status=201, headers=headers)

    def update_member(self):
        id = self.request.matchdict['id']
        member = self.context.update_member(id, self._get_data())
        if member is None:
            member = self.context.create_member(self._get_data())
            headers = {'Location': self.request.path}
            return Response(status=201, headers=headers)
        else:
            return Response(status=204, content_type='')

    def delete_member(self):
        id = self.request.matchdict['id']
        self.context.delete_member(id)
        return Response(status=204, content_type='')

    def render_to_response(self, value, fields=None):
        if value is None:
            raise HTTPNotFound(self.context)
        renderer = self.determine_renderer()
        try:
            renderer = getattr(self, 'render_{0}'.format(renderer))
        except AttributeError:
            name = self.__class__.__name__
            raise HTTPBadRequest(
                '{0} view has no renderer "{1}".'.format(name, renderer))
        return Response(**renderer(value))

    def determine_renderer(self):
        request = self.request
        renderer = (request.matchdict or {}).get('renderer', '').lstrip('.')
        if renderer:
            return renderer
        if request.accept.best_match(['application/json']):
            return 'json'
        elif request.accept.best_match(['application/xml']):
            return 'xml'

    def render_json(self, value):
        response_data = dict(
            body=self.context.to_json(value, self.fields, self.wrap),
            content_type='application/json',
        )
        return response_data

    def render_xml(self, value):
        raise HTTPBadRequest('XML renderer not implemented.')

    @reify
    def fields(self):
        fields = self.request.params.get('$fields', None)
        if fields is not None:
            fields = json.loads(fields)
        return fields

    @reify
    def wrap(self):
        wrap = self.request.params.get('$wrap', 'true').strip().lower()
        return wrap in ('1', 'true')



from setuptools import setup, find_packages


version = open('version.txt').read().strip()


setup(
    name='pyramid_restler',
    version=version,
    description='RESTful views for Pyramid',
    author='Wyatt Lee Baldwin',
    author_email='wyatt.lee.baldwin@gmail.com',
    install_requires=(
        'pyramid>=1.1',
    ),
    packages=find_packages(),
    classifiers=(
        'Development Status :: 2 - Pre-Alpha',
        'Framework :: Pylons',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
    ),
)



from collections import Iterable
import datetime
import decimal
import json

from pyramid.decorator import reify

from sqlalchemy.schema import Column
from sqlalchemy.util import NamedTuple

from zope.interface import implements

from pyramid_restler.interfaces import IContext


datetime_types = (datetime.time, datetime.date, datetime.datetime)


class DefaultJSONEncoder(json.JSONEncoder):

    def default(self, obj):
        """Convert ``obj`` to something JSON encoder can handle."""
        if isinstance(obj, NamedTuple):
            obj = dict((k, getattr(obj, k)) for k in obj.keys())
        elif isinstance(obj, decimal.Decimal):
            obj = str(obj)
        elif isinstance(obj, datetime_types):
            obj = str(obj)
        return obj


class SQLAlchemyORMContext(object):
    """Adapts a SQLAlchemy ORM class to the
    :class:`pyramid_restler.interfaces.IContext` interface."""

    implements(IContext)

    json_encoder = DefaultJSONEncoder

    def __init__(self, request):
        self.request = request

    @reify
    def session(self):
        return self.session_factory()

    def session_factory(self):
        return self.request.db_session

    def get_collection(self):
        q = self.session.query(self.entity)
        return q.all()

    def get_member(self, id):
        q = self.session.query(self.entity)
        return q.get(id)

    def create_member(self, data):
        member = self.entity(**data)
        self.session.add(member)
        self.session.commit()
        return member

    def update_member(self, id, data):
        q = self.session.query(self.entity)
        member = q.get(id)
        if member is None:
            return None
        for name in data:
            setattr(member, name, data[name])
        self.session.commit()
        return member

    def delete_member(self, id):
        q = self.session.query(self.entity)
        member = q.get(id)
        if member is None:
            return None
        self.session.delete(member)
        self.session.commit()

    def get_member_id(self, member):
        pk = member._sa_instance_state.key
        if pk is None:
            return None
        vals = pk[1]
        if len(vals) == 1:
            return vals[0]
        else:
            return tuple(vals)

    def get_member_id_as_string(self, member):
        id = self.get_member_id(member)
        if isinstance(id, basestring):
            return id
        else:
            return json.dumps(id, cls=self.json_encoder)

    def to_json(self, value, fields=None, wrap=True):
        """Convert instance or sequence of instances to JSON.

        ``value`` is a single ORM instance or an iterable that yields
        instances.

        ``fields`` is a list of fields to include for each instance.

        ``wrap`` indicates whether or not the result should be wrapped or
        returned as-is.

        """
        if fields is None:
            fields = self.default_fields
        if isinstance(value, Iterable):
            result = [self.member_to_dict(m, fields) for m in value]
            result_count = len(result)
        else:
            result = self.member_to_dict(value, fields)
            result_count = 1
        if wrap:
            result = dict(
                results=result,
                result_count=result_count,
            )
        return json.dumps(result, cls=self.json_encoder)

    def member_to_dict(self, member, fields=None):
        if fields is None:
            fields = self.default_fields
        return dict((name, getattr(member, name)) for name in fields)

    @reify
    def default_fields(self):
        fields = []
        class_attrs = dir(self.entity)
        for name in class_attrs:
            if name.startswith('_'):
                continue
            attr = getattr(self.entity, name)
            if isinstance(attr, property):
                fields.append(name)
            else:
                try:
                    clause_el = attr.__clause_element__()
                except AttributeError:
                    pass
                else:
                    if issubclass(clause_el.__class__, Column):
                        fields.append(name)
        fields = set(fields)
        return fields
